/* ico2xpm.c -- Convert icons to pixmaps
 * Copyright (C) 1998 Philippe Martin
 * Modified by Brion Vibber
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>

#ifdef STDC_HEADERS
#include <stdlib.h>
#endif

#include "getopt.h"


#ifndef MAX_PATH_LEN
#define MAX_PATH_LEN 1024
#endif

static struct option longopts[] =
{
  { "version", no_argument, NULL, 'v' },
  { "help", no_argument, NULL, 'h' },
  { "output", required_argument, NULL, 'o' },
  { "verbose", no_argument, NULL, 'V' },
  { NULL, 0, NULL, 0 }
};

/* The maximum number of icons we'll try to handle in any one file */
#define MAXICONS 32

/* Header of a single icon image in an icon file */
typedef struct IconDirectoryEntry {
	unsigned char  bWidth;
	unsigned char  bHeight;
	unsigned char  bColorCount;
	unsigned char  bReserved;
	unsigned short  wPlanes;
	unsigned short  wBitCount;
	unsigned long  dwBytesInRes;
	unsigned long  dwImageOffset;
} ICONDIRENTRY;

/* Header of an icon file */
typedef struct ICONDIR {
	 unsigned short          idReserved;
	 unsigned short          idType;
	 unsigned short          idCount;
/*    ICONDIRENTRY  idEntries[1];*/
	 ICONDIRENTRY  idEntries[MAXICONS];
} ICONHEADER;

/* Bitmap header - this is on the images themselves */
typedef struct tagBITMAPINFOHEADER{
        unsigned long      biSize;
        long       biWidth;
        long       biHeight;
        unsigned short       biPlanes;
        unsigned short       biBitCount;
        unsigned long      biCompression;
        unsigned long      biSizeImage;
        long       biXPelsPerMeter;
        long       biYPelsPerMeter;
        unsigned long      biClrUsed;
        unsigned long      biClrImportant;
} BITMAPINFOHEADER;

/* Magic number for an icon file */
/* This is the the idReserved and idType fields of the header */
static char ico_magic_number[4] = {0, 0, 1, 0};

static char *program_name;
static char output_file [MAX_PATH_LEN];
static char input_file [MAX_PATH_LEN];

static int do_verbose = 0;
static int output_given = 0;

static void usage __P((FILE *, int));
static void convert __P((void));
static void get_output_name_from_input_name __P((void));
static void test_different_files __P((void));

/* For keeping track of which icon image we're on */
static int whichimage = 0;

int
main (int argc, char **argv)
{
  int opt;

  program_name = argv[0];

  while ((opt = getopt_long (argc, argv, "hvo:", longopts, (int *) NULL)) != EOF)
    {
      switch (opt)
	{
	case 'h':
	  printf ("Convert MS Windows icons (.ico) to X pixmaps (.xpm)\n");
	  usage (stdout, 0);

	case 'v':
	  printf ("%s %s\n", PACKAGE, VERSION);
	  exit (0);

	case 'o':
	  strcpy (output_file, optarg);
	  output_given = 1;
	  break;

	case 'V':
	  do_verbose = 1;
	  break;

	case 0:
	  break;

	default:
	  usage (stderr, 1);
	}
    }
  
  switch (argc - optind)
    {
    case 0:
      usage (stderr, 1);

    case 1:
      strcpy (input_file, argv[optind]);
      if (!output_given)
	get_output_name_from_input_name ();
      else 
	test_different_files ();
      convert ();
      break;
      
    default:
      if (output_given)
	{
	  fprintf (stderr, 
		   "%s: Warning: you used -o option, only first file will be converted\n",
		   program_name);
	  strcpy (input_file, argv[optind]);
	  test_different_files ();
	  convert ();
	}
      else
	while (argv[optind])
	  {
	    strcpy (input_file, argv[optind]);
	    get_output_name_from_input_name ();
	    convert ();
	    optind ++;
	  }
    }
  return 0;
}


static void
usage (FILE *f, int status)
{
  fprintf (f, "Usage: %s infile ... -o outfile\n", program_name);
  exit (status);
}

static void
get_output_name_from_input_name ()
{
  size_t l = strlen (input_file);

  if (l >= 3
      && input_file [l-4] == '.'
      && (input_file [l-3] == 'i' || input_file [l-3] == 'I')
      && (input_file [l-2] == 'c' || input_file [l-2] == 'C')
      && (input_file [l-1] == 'o' || input_file [l-1] == 'O'))
    {
      strncpy (output_file, input_file, l - 4);
      output_file[l-4] = '\0';
    }
  else
    strcpy (output_file, input_file);

  if(whichimage > 0) {
    char n[4];
    sprintf(n, "-%d", whichimage+1);
    strcat (output_file, n);
  }

  strcat (output_file, ".xpm");
}

static void
test_different_files (void)
{
  if (strcmp (input_file, output_file) == 0)
    {
      fprintf (stderr, "%s: input and output files are the same: %s\n", 
	       program_name, input_file);
      exit (1);
    }
}

/*
 * Conversion
 */

#define XPM_HEAD "/* XPM */\n\
static char *pixmap[] = {\n\
/* width height ncols cpp */\n\
\"%d %d %d 2\",\n  /* Colors */\n"

/* The smallest number of double words containing c bytes */
#define DWORD_ALIGN_BYTES(b) (((b) / 4) + (((b) % 4 > 0) ? 1 : 0))

/* The smallest number of double words containing c bits */
#define DWORD_ALIGN_BITS(b) (((b) / 32) + (((b) % 32 > 0) ? 1 : 0))

/* The different possible number of colors */
#define LOW_COLOR 16
#define HIGH_COLOR 0
/*#define NCOLORS (color_level == LOW_COLOR ? 16 : 256)*/
#define NCOLORS ncolors

/* The color index of a pixel in the icon colormap */
#define PIXEL_INDEX(x,y) \
     (color_level == LOW_COLOR \
      ? ((x) % 2 == 0 \
	 ? (image [bytes_per_line * (height - 1 - (y)) / 2 + (x) / 2] & 0xF0) >> 4 \
	 : image [bytes_per_line * (height - 1 - (y)) / 2 + (x) / 2] & 0x0F) \
      : image [bytes_per_line * (height - 1 - (y)) + (x)])
     
/* The color index of a pixel in the (reduced) pixmap colormap */
#define PIXEL_REDUCED_INDEX(x,y) \
     (reduced_colormap_index [PIXEL_INDEX((x),(y))])
  
/* The RGB value of a colormap entry */
#define INDEX_VALUE(index,c) (colormap[(c) + (index) * 4])
#define INDEX_R_VALUE(index) (INDEX_VALUE((index), 2))
#define INDEX_G_VALUE(index) (INDEX_VALUE((index), 1))
#define INDEX_B_VALUE(index) (INDEX_VALUE((index), 0))
  
/* The transparency of a pixel */
#define MASK_BYTE(x,y) (bytes_per_mask_line * (height - 1 - (y)) + (x) / 8)
#define MASK_BIT(x,y) (7 - (x) % 8)
#define IS_TRANSPARENT(x,y) \
     (mask[MASK_BYTE((x),(y))] & (1 << MASK_BIT((x),(y))))
  
  
static FILE *ico_stream;
static FILE *xpm_stream;

/* Values read from icon file */
static ICONHEADER iconheader;
static BITMAPINFOHEADER bitmapinfoheader;
/*static unsigned char magic_number [sizeof(ico_magic_number)];*/
static unsigned char *magic_number;
static unsigned char cwidth;
static unsigned char cheight;
static unsigned char color_level;
static unsigned char unknown_part [53];
static unsigned char colormap [256 * 4];
static unsigned char image [256 * 256]; 
static unsigned char mask [256 * 32];

/* Different variables computed from the read ones */
static int ncolors;
static int width;
static int height;
static unsigned char bytes_per_line;
static unsigned char bytes_per_mask_line;
static int image_length;
static int mask_length;
static int i, x, y;

/* Variables for the reduced colormap */
static int color_used[256];
static int nb_color_used;
static unsigned char reduced_colormap_index [256];

/* Does the icon have transparent pixels ? */
static int have_transparent_pixels;


static void
convert (void)
{
  /* 
   * Read the icon file 
   */
  
  if ((ico_stream = fopen (input_file, "r")) == NULL)
    {
      fprintf (stderr, "%s: ", program_name);
      perror (input_file);
      return;
    }

  /* Check the magic number & header */
  /*fread (magic_number, 1, sizeof(ico_magic_number), ico_stream);*/
  fread (&iconheader, 1, 6/*sizeof(unsigned short)*3*/, ico_stream);
  magic_number = (unsigned char *)&iconheader;
  for (i = 0 ; i < sizeof(ico_magic_number) ; i ++)
    {
      if (magic_number [i] != ico_magic_number [i])
	{
	  fprintf (stderr, "%s: %s: Not a recognized icon file.\n", 
		   program_name, input_file);
	  fclose (ico_stream);
	  return;
	}
    }
  
  /* Output some stats */
  if (do_verbose)
    fprintf (stdout, "%s: %s: file contains %d icon images\n",
               program_name, input_file, iconheader.idCount);

  /* Read in the rest of the icon directory entries */
  fread(&iconheader.idEntries[0],iconheader.idCount,
    /*sizeof(ICONDIRENTRY)*/16,ico_stream);
  
  /* Cycle through each icon image */
  for(whichimage = 0; whichimage < iconheader.idCount; whichimage++) {
    /* For debugging
    if(do_verbose) {
      fprintf(stdout, "(%d) identry: %d %d %d %d %d %d %d %d\n",
        whichimage,
        (int)iconheader.idEntries[whichimage].bWidth,
        (int)iconheader.idEntries[whichimage].bHeight,
        (int)iconheader.idEntries[whichimage].bColorCount,
        (int)iconheader.idEntries[whichimage].bReserved,
        (int)iconheader.idEntries[whichimage].wPlanes,
        (int)iconheader.idEntries[whichimage].wBitCount,
        (int)iconheader.idEntries[whichimage].dwBytesInRes,
        (int)iconheader.idEntries[whichimage].dwImageOffset);
    }
    */
    
    /* Update the output file name */
    get_output_name_from_input_name();

    /* Read the icon size */
    width = iconheader.idEntries[whichimage].bWidth;
    height = iconheader.idEntries[whichimage].bWidth;

    /* Determine the number of bytes defining a line of the icon. */
    bytes_per_line = 4 * DWORD_ALIGN_BYTES (width);
    bytes_per_mask_line = 4 * DWORD_ALIGN_BITS (width);

    /* Let's surf on over to the bitmap image & read the BMIH. */
    fseek (ico_stream, iconheader.idEntries[whichimage].dwImageOffset,SEEK_SET);
    fread (&bitmapinfoheader, 1, sizeof(bitmapinfoheader), ico_stream);

    /* For debugging
    if(do_verbose) {
      fprintf(stdout,"(%d) bmih: %d %d %d %d %d %d %d %d %d %d %d\n",
        whichimage,
        (int)bitmapinfoheader.biSize,
        (int)bitmapinfoheader.biWidth,
        (int)bitmapinfoheader.biHeight,
        (int)bitmapinfoheader.biPlanes,
        (int)bitmapinfoheader.biBitCount,
        (int)bitmapinfoheader.biCompression,
        (int)bitmapinfoheader.biSizeImage,
        (int)bitmapinfoheader.biXPelsPerMeter,
        (int)bitmapinfoheader.biYPelsPerMeter,
        (int)bitmapinfoheader.biClrUsed,
        (int)bitmapinfoheader.biClrImportant);
    }*/
    
    /* Read the number of colors.
     * TODO: add support for monochrome, 24-bit icons
     */
    switch(bitmapinfoheader.biPlanes
        * bitmapinfoheader.biBitCount) {
      case 4: /* 2^4 = 14 */ color_level = LOW_COLOR; break;
      case 8: /* 2^8 = 256 */ color_level = HIGH_COLOR; break;
      default:
        fprintf(stderr,
          "%s: %s: Unsupported number of colors in image! Skipping. (%d planes, %d bpp)\n",
          program_name, input_file,
          (int)bitmapinfoheader.biPlanes,
          (int)bitmapinfoheader.biBitCount);
        continue;
    }
    
    /* Read the colormap */
    if (bitmapinfoheader.biClrImportant != 0)
      ncolors = bitmapinfoheader.biClrImportant;
    else if (bitmapinfoheader.biClrUsed != 0)
      ncolors = bitmapinfoheader.biClrUsed;
    else
      ncolors = 1 << (bitmapinfoheader.biPlanes*bitmapinfoheader.biBitCount);

    fseek (ico_stream, iconheader.idEntries[whichimage].dwImageOffset
      + bitmapinfoheader.biSize, SEEK_SET);
    fread (colormap, 1,ncolors * 4, ico_stream);

    /* Read the image */
    if (color_level == LOW_COLOR)
        image_length = bytes_per_line * height / 2;
    else
        image_length = bytes_per_line * height;
    fread (image, 1, image_length, ico_stream);

    /* Read the mask */
    mask_length = bytes_per_mask_line * height;
    fread (mask, 1, mask_length, ico_stream);

    /* Read error / prematured EOF  occured ? */
    if (feof (ico_stream))
      fprintf (stderr, "%s: %s: Premature end-of-file.\n",
	     program_name, input_file);
    if (ferror (ico_stream))
      fprintf (stderr, "%s: %s: Read error.\n",
  	     program_name, input_file);
    if (feof (ico_stream) || ferror (ico_stream))
      {
        fclose (ico_stream);
        return;
      }

    /* Reduce the number of colors */
    for (i = 0 ; i < ncolors ; i++)
      color_used [i] = 0;
  
    for (y = 0 ; y <height ; y++)
      for (x = 0 ; x < width ; x++)
        color_used [PIXEL_INDEX(x,y)] = 1;
  
    nb_color_used = 0;
    for (i = 0 ; i < ncolors ; i++)
      if (color_used [i])
        reduced_colormap_index[i] = nb_color_used ++;

    /* Check for at least one transparent pixel */
    have_transparent_pixels = 0;
    for (y=0; y<height && have_transparent_pixels == 0 ; y++)
      for (x=0; x<width; x++)
        if (IS_TRANSPARENT(x,y))
	  {
	    have_transparent_pixels = 1;
	    break;
	  }
  
    /*
     * Write the pixmap file
     */

    if ((xpm_stream = fopen (output_file, "w")) == NULL)
      {
        fprintf (stderr, "%s: ", program_name);
        perror (output_file);
        return;
      }

    /* Write the header */
    fprintf (xpm_stream, XPM_HEAD, 
	     width, height, 
	     nb_color_used + have_transparent_pixels);

    /* Write the colors */
    x = 0;
    for (i=0; i<NCOLORS; i++)
      {
        if (color_used [i])
	  {
	    fprintf (xpm_stream, "  \"%.2X c #%.2X%.2X%.2X\",\n", x++, 
		    INDEX_R_VALUE(i), INDEX_G_VALUE(i), INDEX_B_VALUE(i));
	  }
      }
    if (have_transparent_pixels)
      fprintf (xpm_stream, "  \".. s None c None\",\n");

    /* Write the image */
    for (y=0; y<height; y++)
      {
        fprintf (xpm_stream, "  \"");
        for (x=0; x<width; x++)
	  {
	    if (IS_TRANSPARENT(x,y))
	      fprintf (xpm_stream, "..");
	    else
	      fprintf (xpm_stream, "%.2X", PIXEL_REDUCED_INDEX(x,y));
	  }
        fprintf (xpm_stream, "\",\n");
      }
    fprintf (xpm_stream, "};\n");

    /* Write error occured ? */
    if (ferror (xpm_stream))
        fprintf (stderr, "%s: %s: Error writing %s.\n",
	         program_name, input_file, output_file);
  
    fclose (xpm_stream);
  
    /* Write info about xpm file written */
    if (do_verbose)
      fprintf (stdout, "%s: %s: Wrote %s %dx%dx%d (%d)\n",
	       program_name, input_file, output_file, width, height,
	       nb_color_used, NCOLORS);
  }
  fclose (ico_stream);
}
